var axios = require("axios");
var Bet = require("./Bet.js");
var User = require("./User.js");
var fs = require("fs").promises;
var file = require("fs");
const { EventEmitter } = require('events');

class UFC extends EventEmitter {

    constructor(users, upComingMatches, previousMatches, outstandingBets, lastRefreshed) {
        if (!users) {
            try {
                // "./node_modules/ufc-betting-game/utils/"
                var jsonUsers = JSON.parse(file.readFileSync(`${__dirname}/../../../users.json`));
                users = jsonUsers;
                console.log("Successfully read users.json for user data.");
            } catch (err) {
                users = [];
            }
        }
        if (!upComingMatches) {
            upComingMatches = [];
        }
        if (!outstandingBets) {
            try {
                var jsonBets = JSON.parse(file.readFileSync(`${__dirname}/../../../bets.json`));
                outstandingBets = jsonBets;
                console.log("Successfully read bets.json for outstandingBets.");
            } catch (err) {
                outstandingBets = [];
            }
        }

        if (!lastRefreshed) lastRefreshed = 0;
        this.users = users; //List of users
        this.upComingMatches = upComingMatches; //List of upcoming matches
        this.previousMatches = previousMatches; //List of previous matches
        this.outstandingBets = outstandingBets; //List of previous matches
        this.lastRefreshed = lastRefreshed; //Ex. Last time the upComingMatches was refreshed.   
        this.checkFiles(); //Just checks to see if all the necessary json files are made.

    }

    /*
    Webscrapes oddshark UFC website for match data.
    Refreshes the match data. Gets all upComingMatches and populates the datasection with those. Then gets all previousMatches and updates the json as well as data section
    */
    async refreshUpComingMatches() {
        try {
            const time = Date.now();
            const options = {
                headers: {
                    'authority': "io.oddsshark.com",
                    'method': "GET",
                    'path': `/ticker/ufc?_=${time}`,
                    'scheme': "https",
                    'accept': "*/*",
                    'accept-encoding': "gzip, deflate, br",
                    'accept-language': "en-US,en;q=0.9,ja;q=0.8",
                    'dnt': "1",
                    'origin': "https://www.oddsshark.com",
                    'referer': "https://www.oddsshark.com/",
                    'sec-fetch-dest': "empty",
                    'sec-fetch-mode': "cors",
                    'sec-fetch-site': "same-site",
                    'user-agent': "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.190 Safari/537.36"
                }
            }
            var response = await axios.get(`https://io.oddsshark.com/ticker/ufc?_=${time}`, options)
            var data = await UFC.parseMatchDataJson(response.data)
            this.upComingMatches = data.upComingMatches;
            this.previousMatches = data.previousMatches;
            this.lastRefreshed = Date.now();
            await fs.writeFile(`${__dirname}/../../../matchData.json`, JSON.stringify(response.data, null, 2));
            await fs.writeFile(`${__dirname}/../../../previousMatches.json`, JSON.stringify(this.previousMatches, null, 2));
            return true;
        } catch (err) {
            console.log(err);
            return false;
        }
    }

    /*
    Uses hardcoded text file paths and checks to see if they exist then creates those files if they dont
    */
    async checkFiles() {
        var files = [`${__dirname}/../../../resolvedBets.json`, `${__dirname}/../../../matchData.json`, `${__dirname}/../../../users.json`, `${__dirname}/../../../previousMatches.json`, `${__dirname}/../../../bets.json`];
        for (var file of files) {
            try {
                await fs.readFile(file);
            } catch (err) {
                console.log(`${file} not found. Generating new blank file.`)
                fs.writeFile(file, "[]")
            }
        }
    }

    async addUser(uuid, name) {
        //If there is a user that exists with this UUID
        if (this.users.find(user => user.uuid == uuid)) {
            console.log("User already exists with this UUID.");
            return false;
        } else {
            this.users.push(new User(uuid, name)); //The rest will be autofilled from the constructor.
            console.log("Adding new user.");
            await this.writeUsersToFile();
            return true;
        }
    }

    async addBet(bet) {
        //------Checks to see if there are any bets of the same type between the SAME 1 or 2 people
        //Checks to see if there are any bets on the same fight between the SAME 1 or 2 people
        let foundBet;
        if (bet.betType == "classic") {
            foundBet = this.outstandingBets.find(b => (b.betType == bet.betType && b.user1.uuid == bet.user1.uuid && b.fightEventID == bet.fightEventID));
            // foundBet = await this.findOutstandingBet(bet.betType, bet.user1.uuid, null);
            // foundBet = await this.findOutstandingBet(bet.betType, bet.user1.uuid, null);
        } else {
            // foundBet = await this.findOutstandingBet(bet.betType, bet.user1.uuid, bet.user2.uuid);
            foundBet = this.outstandingBets.find(b => (b.betType == bet.betType && ((b.user1.uuid == bet.user1.uuid && b.user2.uuid == bet.user2.uuid) || (b.user1.uuid == bet.user2.uuid && b.user2.uuid == bet.user1.uuid)) && b.fightEventID == bet.fightEventID));

        }

        if (foundBet) {
            console.log("Error: bet of this type with these users on this fight already exists.");
            return false;
        } else if (this.previousMatches.find(match => match.event_id == bet.fightEventID)) {
            console.log("Error: tried to make bet on a fight that is already over");
            return false;
        } else {
            try {
                let user1 = await this.findUser(bet.user1.uuid);
                console.log("Adding new bet")
                user1.currentBets.push(bet);
                if (bet.user2) {
                    let user2 = await this.findUser(bet.user2.uuid);
                    if (user2) {
                        user2.currentBets.push(bet);
                    }
                }

                this.outstandingBets.push(bet);
                await this.writeBetsToFile();
                await this.writeUsersToFile();
                return true;
            } catch (err) {
                console.log(err);
                console.log("Failed to add to outstandingBets")
                return false;
            }

        }
    }

    /*
    Finds the outstanding bet that you want and deletes that bet then writes it to file to ensure its definitely gone.
    */
    async cancelBet(betType, fightID, user1ID, user2ID) {
        try {
            let bet = await this.findOutstandingBetWithFightID(betType, fightID, user1ID, user2ID);
            if (bet) {
                this.outstandingBets.splice(this.outstandingBets.indexOf(bet), 1);
                let user1 = await this.findUser(bet.user1.uuid);
                user1.currentBets.splice(user1.currentBets.findIndex(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid), 1);
                try {
                    let user2 = await this.findUser(bet.user2.uuid);
                    user2.currentBets.splice(user2.currentBets.findIndex(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid), 1);
                } catch (err) {
                }
                await this.writeBetsToFile();
                await this.writeUsersToFile();
                return true;
            } else {
                console.log("Bet to delete wasn't found.");
                return false;
            }
        } catch (err) {
            console.log(err)
            return false;
        }

    }

    /* 
    Utility function to find certain outstanding bets.
    */
    async findOutstandingBet(betType, user1ID, user2ID) {
        var foundBet = null;
        if (betType == "classic") {
            foundBet = this.outstandingBets.find(b => (b.betType == betType && b.user1.uuid == user1ID));
        } else if (betType == "1v1") {
            foundBet = this.outstandingBets.find(b => b.betType == betType && ((b.user1.uuid == user1ID && b.user2.uuid == user2ID) || b.user1.uuid == user2ID && b.user2.uuid == user1ID));
        }
        if (foundBet) {
            return foundBet;
        } else
            return null;
    }


    /* 
    Utility function to find certain outstanding bets.
    */
    async findOutstandingBetWithFightID(betType, fightID, user1ID, user2ID) {
        var foundBet = null;
        if (betType == "classic") {
            foundBet = this.outstandingBets.find(b => (b.betType == betType && b.user1.uuid == user1ID) && b.fightEventID == fightID);
        } else if (betType == "1v1") {
            foundBet = this.outstandingBets.find(b => b.betType == betType && ((b.user1.uuid == user1ID && b.user2.uuid == user2ID) || b.user1.uuid == user2ID && b.user2.uuid == user1ID) && b.fightEventID == fightID);
        }
        if (foundBet) {
            return foundBet;
        } else
            return null;
    }


    async verify1v1Bet(betType, fightID, user1ID, user2ID) {
        try {
            let bet = await this.findOutstandingBetWithFightID(betType, fightID, user1ID, user2ID);
            if (bet && betType == "1v1") {
                let user1 = await this.findUser(user1ID);
                let user2 = await this.findUser(user2ID);
                let user1Bet = user1.currentBets.find(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid && b.user2.uuid == bet.user2.uuid);
                let user2Bet = user2.currentBets.find(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid && b.user2.uuid == bet.user2.uuid);
                user1Bet.accepted = true;
                user2Bet.accepted = true;
                bet.accepted = true;
                await this.writeBetsToFile();
                await this.writeUsersToFile();
                return bet;
            }
        } catch (err) {
            return false
        }
        return false;
    }
    /*
    Loops through all existing bets within outstandingBets file to find all the bets that SHOULD be ready to be completed. If the fight is within previousMatches (over
    and decision exists), resolves the bet from there. If not, doesnt do anything.
    */
    async resolveBets() {
        var resolvedBets = [];
        console.log('resolve bets');
        for (var bet of this.outstandingBets) {
            if (Date.now() > bet.fightEventDate) { //If we passed the bets fight date. Then we want to check the completion of the fight
                var fight = this.previousMatches.find(m => m.event_id == bet.fightEventID);
                if (fight) {
                    var winnerID = null;
                    var loserID = null;
                    switch (bet.betType) {
                        case "classic":

                            //If the user won. Reference https://www.gamingtoday.com/tools/moneyline/ for calculating winnings
                            if (bet.user1.fighterName == fight.winner) {
                                var cashWon = 0;
                                winnerID = bet.user1.uuid;
                                //Must use odds saved in the bet data. Sometimes, odds will change, so if we scrape website again, it will have different odds.
                                if (bet.odds.user1 > 0) cashWon = (bet.betAmount * bet.odds.user1 / 100);
                                else if (bet.odds.user1 < 0) cashWon = (bet.betAmount / (-1 * bet.odds.user1 / 100));

                                //now add cashWon + betAmount to the users account.
                                console.log("classic won: " + cashWon);
                                await this.addMoney(winnerID, parseInt(cashWon + bet.betAmount));
                                this.emit('betResolved', bet, "WON", winnerID, cashWon + bet.betAmount);
                                //If the user lost. Don't give any money. We've already taken money from their account
                            } else if (fight.winner != "") {
                                loserID = bet.user1.uuid;
                                console.log("classic lose");
                                this.emit('betResolved', bet, "LOST", null, 0);

                                //Match was a draw. No one wins. Give back money
                            } else {
                                //Give back betAmount to the user.
                                console.log("classic draw");
                                await this.addMoney(bet.user1.uuid, bet.betAmount);
                                this.emit('betResolved', bet, "DRAW", null, 0);
                            }

                            break;
                        case "1v1":
                            //If there was a winner
                            if (fight.winner != "") {
                                winnerID = [bet.user1, bet.user2].find(user => user.fighterName == fight.winner);
                                loserID = [bet.user1, bet.user2].find(user => user != winnerID);

                                //Give winner bet.betAmount * 2;
                                await this.addMoney(winnerID, (bet.betAmount * 2))
                                this.emit('betResolved', bet, "WON", winnerID, bet.betAmount * 2);

                                //Otherwise, its a draw. No one wins. Give back both their money
                            } else {
                                await this.addMoney(bet.user1.uuid, bet.betAmount)
                                await this.addMoney(bet.user2.uuid, bet.betAmount)
                                this.emit('betResolved', bet, "DRAW", null, 0);
                            }

                            break;
                        default:
                            break;
                    }
                    //After all betting is done and over, remove the bet from the list and add it to resolvedBets json.
                    bet.betResolved = true;
                    resolvedBets.push(bet);
                    //Removes bet from outstandingBets
                    this.outstandingBets.splice(this.outstandingBets.indexOf(bet), 1);
                    let user1 = await this.findUser(bet.user1.uuid);
                    user1.currentBets.splice(user1.currentBets.findIndex(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid), 1);
                    try {
                        let user2 = await this.findUser(bet.user2.uuid);
                        user2.currentBets.splice(user2.currentBets.findIndex(b => b.betType == bet.betType && b.fightEventID == bet.fightEventID && b.user1.uuid == bet.user1.uuid), 1);
                    } catch (err) {
                    }


                }

            }
        }
        //Want to write all this to file since weve updated and resolved alot of bets;
        try {
            // console.log(resolvedBets.length);
            if (resolvedBets.length > 0) {
                var jsonResolvedBets = JSON.parse(await fs.readFile(`${__dirname}/../../../resolvedBets.json`));
                jsonResolvedBets = jsonResolvedBets.concat(resolvedBets);
                await fs.writeFile(`${__dirname}/../../../resolvedBets.json`, JSON.stringify(jsonResolvedBets, null, 2))
                await this.writeBetsToFile();
                await this.writeUsersToFile();
            }
            return true;
        } catch (err) {
            console.log(err);
            console.log("Error saving jsonResolvedBets/Users to the json file.")
            return false;

        }
    }

    async addMoney(uuid, moneyGiven) {
        try {
            moneyGiven = parseInt(moneyGiven);
            var user = this.users.find(user => user.uuid == uuid);
            let newBalance = parseInt(user.balance) + moneyGiven;
            if (isNaN(newBalance)) throw new Error("New balance wasn't a real number")
            user.balance = newBalance;
            await this.writeUsersToFile();
            return true;

        } catch (err) {
            // console.log(err)
            return false;
        }
    }


    async loadFromFile(filePath) {
        try {
            if (!filePath) filePath = `${__dirname}/../../../matchData.json`;
            var matchData = JSON.parse(file.readFileSync(filePath));
            var data = await UFC.parseMatchDataJson(matchData);
            this.upComingMatches = data.upComingMatches;
            this.previousMatches = data.previousMatches;
            this.lastRefreshed = Date.now();
            console.log("Successfully read matchData.json for upComingMatches and previousMatches.");
            await fs.writeFile(`${__dirname}/../../../previousMatches.json`, JSON.stringify(this.previousMatches, null, 2));
            return true;
        } catch (err) {
            console.log(err);
            return false;
        }
    }

    /*
    Simple function to remove money from a Users account and write that to file.
    */
    async takeMoney(uuid, moneyTaken) {
        try {
            moneyTaken = parseInt(moneyTaken);
            var user = this.users.find(user => user.uuid == uuid);
            let newBalance = parseInt(user.balance) - moneyTaken;
            if (parseInt(user.balance) - moneyTaken < 0) throw new Error("User balance can't fall below 0");
            if (isNaN(newBalance)) throw new Error("User balance is not a real number.")
            user.balance = newBalance
            await this.writeUsersToFile();
            return true;
        } catch (err) {
            // console.log(err)
            return false;
        }
    }


    /*
    Function that saves our outstandingBets datasection to file so that we can read from file anytime we boot.
    This function should get called anytime we make changes to the outstandingBets.
    */
    async writeBetsToFile() {
        console.log("Write Bets To File");
        try {
            await fs.writeFile(`${__dirname}/../../../bets.json`, JSON.stringify(this.outstandingBets, null, 2));
            return true;
        } catch (err) {
            console.log(err);
            return false;
        }
    }

    async writeUsersToFile() {
        console.log("Write Users To File");
        try {
            await fs.writeFile(`${__dirname}/../../../users.json`, JSON.stringify(this.users, null, 2));
            return true;
        } catch (err) {
            console.log(err);
            return false;
        }
    }

    async findUser(uuid) {
        try {
            return this.users.find(user => user.uuid == uuid);
        } catch (err) {
            console.log(err);
            return null;
        }
    }

    /*
    Retrieves a fight depending the fightEventID, which is the unique ID given to every fight.
    */
    async getFight(fightEventID) {
        try {
            return this.upComingMatches.find(match => match.event_id == fightEventID);
        } catch (err) {
            console.log(err);
            return null;
        }
    }

    /*
    Parses the matchData from the UFC json data that we receive. return it as {upcomingMatches, previousMatches} but previousMatches has ALL past saved matches.
    */
    static async parseMatchDataJson(data) {
        try {
            var upComingMatches = [];
            var previousMatches = JSON.parse(await fs.readFile(`${__dirname}/../../../previousMatches.json`));
            if (data.matchups) {
                for (var fight of data.matchups) {
                    try {


                        if (fight.type == "matchup") {
                            //Starts by fixing unformated dates. Puts them in milliseconds form for easier use.
                            var date = Date.parse(fight.event_date);
                            if (isNaN(date)) date = Date.now();
                            fight.event_date = date;
                            if (fight.status == "") {
                                if (!isNaN(fight.away_odds) && !isNaN(fight.home_odds)) {
                                    upComingMatches.push(fight);
                                }
                            } else {
                                if (!previousMatches.find(match => match.event_id == fight.event_id))
                                    previousMatches.push(fight);

                            }
                        }
                    } catch (err) {
                        console.log(err);
                        continue;
                    }
                }
            }
            return { upComingMatches: upComingMatches, previousMatches: previousMatches }
        } catch (err) {
            console.log(err);
            return null;
        }
    }
}
main();
/*
Test function
*/
async function main() {
    // var test = new UFC();
    // await test.loadFromFile();
    // await test.refreshUpComingMatches();
    // await test.resolveBets();

    // console.log(test.previousMatches)
    // console.log(test.upComingMatches.length);
    // console.log(test.previousMatches.length);
    // var john = await test.findUser(1234);
    // var bob = await test.findUser(456);
    // var fight = test.getFight()
    // test.addUser(123, "john");
    // test.addUser(456, "bob");
    // console.log(test.outstandingBets);
    // test.takeMoney(456, 777)
    // test.addMoney(123, 4);
    // await test.refreshUpComingMatches();
    // console.log(test.upComingMatches);
    // console.log(test.previousMatches);
    // await test.addBet(new Bet("1v1", 200, 1382448, 1615694400000, {user: john, fighterName:"M Nicolau"}, {user: bob, fighterName:"T Ulanbekov"}, {user1: "125", user2: "-145"} ))
    // await test.addBet(new Bet("classic", 300, 1370716, 1615096800000, {user: john, fighterName:"I Adesanya"}, null, {user1: "-250", user2: null} ))
    // await test.resolveBets();
    // await test.cancelBet("classic", john, null);
    // console.log(test.outstandingBets);
}


module.exports = UFC;